#!/usr/bin/env python

"""
Transform Neurokernel-compatible NetworkX graph to/from Neuroarch-compatible graph.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import copy
import itertools
import numbers

import networkx as nx

def na_pat_to_nk(g):
    """
    Transform a graph containing a Neuroarch-compatible pattern into
    one that can be executed by Neurokernel.

    Parameters
    ----------
    g : networkx.MultiDiGraph
        Graph containing Neuroarch-compatible pattern.

    Returns
    -------
    g : networkx.MultiDiGraph
        Graph containing Neurokernel-executable pattern.
    """

    assert isinstance(g, nx.MultiDiGraph)

    # Find interfaces:
    g_new = nx.MultiDiGraph()
    int_id_to_name = {}
    for n, data in g.nodes(data=True):
        if data['class'] == 'Interface':
            int_id_to_name[n] = data['name']

    # Create ports in new graph:
    old_port_id_to_new = {}
    for int_id in int_id_to_name:
        for old_port_id in g.successors(int_id):
            data = g.nodes[old_port_id]
            new_port_id = data['selector']
            g_new.add_node(new_port_id,
                           **{'interface': int_id_to_name[int_id],
                                      'io': data['port_io'],
                                      'type': data['port_type']})
            old_port_id_to_new[old_port_id] = new_port_id

    # Create connections between ports in the two interfaces:
    for from_id, to_id, data in g.edges(data=True):
        if data['class'] == 'SendsTo' and \
           g.nodes[from_id]['class'] == 'Port' and \
           g.nodes[to_id]['class'] == 'Port':
            g_new.add_edge(old_port_id_to_new[from_id],
                           old_port_id_to_new[to_id])

    return g_new

def na_lpu_to_nk(g):
    """
    Transform a graph containing a Neuroarch-compatible circuit into
    one that can be executed by Neurokernel.

    Parameters
    ----------
    g : networkx.MultiDiGraph
        Graph containing Neuroarch-compatible circuit.

    Returns
    -------
    g : networkx.MultiDiGraph
        Graph containing Neurokernel-executable circuit.
    """

    assert isinstance(g, nx.MultiDiGraph)

    g_new = nx.MultiDiGraph()
    # id_to_label = {}

    for id, data in g.nodes(data=True):
        if data['class'] in ['Interface', 'LPU']:
            continue

        # Don't clobber the original graph's data:
        data = copy.deepcopy(data)
        # g_new.add_node(data['label'], **data)
        # id_to_label[id] = data['label']
        g_new.add_node(id, **data)

    # Create synapse edges:
    for from_id, to_id, data in g.edges(data = True):
        data = copy.deepcopy(data)
        if data.pop('class') == 'SendsTo':
            # g_new.add_edge(id_to_label[from_id], id_to_label[to_id], **data)
            g_new.add_edge(from_id, to_id, **data)

    return g_new

na_lpu_to_nk_new = na_lpu_to_nk


from unittest import main, TestCase

import neuroarch.conv.pd
import deepdiff
import pyorient
import pandas as pd

db_name = 'neuroarch_test_db'
username = 'admin'
passwd = 'admin'

match = lambda a, b: False if deepdiff.DeepDiff(a, b) else True

class TestConvPandas(TestCase):    
    @classmethod
    def setUpClass(cls):
        cls.client = pyorient.OrientDB('localhost', 2424)
        cls.client.connect(username, passwd)
        if cls.client.db_exists(db_name):
            cls.client.db_drop(db_name)
        cls.client.db_create(db_name, pyorient.DB_TYPE_GRAPH,
                             pyorient.STORAGE_TYPE_MEMORY)
        cls.client.db_open(db_name, username, passwd)

    @classmethod
    def tearDownClass(cls):
        cls.client.connect(username, passwd)
        try:
            cls.client.db_drop(db_name)
        except Exception as e:
            warnings.warn('problem cleaning up test database: %s' % e.message)

    def _connect_server(self):
        self.client = pyorient.OrientDB('localhost', 2424)
        self.client.connect(username, passwd)

    def setUp(self):
        cmds = ['create class neuron extends V',
                'create class synapse extends V',
                'create class data extends E']
        for cmd in cmds:
            self.client.command(cmd)

    def tearDown(self):
        cmds = ['delete vertex neuron',
                'delete vertex synapse',
                'delete edge data',
                'drop class neuron',
                'drop class synapse',
                'drop class data']
        for cmd in cmds:
            self.client.command(cmd)

    def _create_pandas_graph(self):
        df_node = pd.DataFrame({'name': ['foo', 'bar', 'baz', 
                                         'foo-bar', 'foo-baz'],
                                'class': ['neuron', 'neuron', 'neuron',
                                          'synapse', 'synapse']})
        df_edge = pd.DataFrame({'out': [0, 3, 0, 4],
                                'in': [3, 1, 4, 2],
                                'class': ['data', 'data', 'data', 'data']})
        return df_node, df_edge

    def _create_orient_graph(self):
        cmd = ("begin;"               
               "let foo = create vertex neuron content {'name': 'foo', 'id': 0};"
               "let bar = create vertex neuron content {'name': 'bar', 'id': 1};"
               "let baz = create vertex neuron content {'name': 'baz', 'id': 2};"
               "let foo_bar = create vertex synapse content {'name': 'foo-bar', 'id': 3};"
               "let foo_baz = create vertex synapse content {'name': 'foo-baz', 'id': 4};"
               "create edge data from $foo to $foo_bar;"
               "create edge data from $foo_bar to $bar;"
               "create edge data from $foo to $foo_baz;"
               "create edge data from $foo_baz to $baz;"
               "commit retry 5;")
        self.client.batch(cmd)

    def test_orient_to_pandas(self):
        df_node_pandas, df_edge_pandas = self._create_pandas_graph()
        self._create_orient_graph()
        df_node_orient, df_edge_orient = neuroarch.conv.pd.orient_to_pandas(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        self.assertSetEqual(set([tuple(v) for v in df_node_pandas.values]),
                            set([tuple(v) for v in df_node_orient.values]))
        self.assertSetEqual(set([tuple(v) for v in df_edge_pandas.values]),
                            set([tuple(v) for v in df_edge_orient.values]))
        self.assertSetEqual(set(df_node_pandas.index),
                            set(df_node_orient.index))

    def test_pandas_to_orient(self):
        df_node_pandas, df_edge_pandas = self._create_pandas_graph()
        neuroarch.conv.pd.pandas_to_orient(self.client,
                                           df_node_pandas, df_edge_pandas)
        df_node_orient, df_edge_orient = neuroarch.conv.pd.orient_to_pandas(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        self.assertSetEqual(set([tuple(v) for v in df_node_pandas.values]),
                            set([tuple(v) for v in df_node_orient.values]))
        self.assertSetEqual(set([tuple(v) for v in df_edge_pandas.values]),
                            set([tuple(v) for v in df_edge_orient.values]))
        self.assertSetEqual(set(df_node_pandas.index),
                            set(df_node_orient.index))

    def test_pandas_to_orient_double(self):
        df_node_pandas = pd.DataFrame({'name': ['foo', 'bar', 'foo-bar'],
                                       'class': ['neuron', 'neuron', 'synapse'],
                                       'x': [1/3.0, 1/4.0, 1.0]})
        df_edge_pandas = pd.DataFrame({'out': [0, 2], 'in': [2, 1],
                                       'class': ['data', 'data']})
        neuroarch.conv.pd.pandas_to_orient(self.client,
                                           df_node_pandas,
                                           df_edge_pandas)
        df_node_orient, df_edge_orient = \
            neuroarch.conv.pd.orient_to_pandas(self.client,
                    'g.V.has("@class", T.in, ["neuron","synapse"])',
                    'g.E.has("@class", "data")')
        self.assertSetEqual(set([tuple(v) for v in df_node_pandas.values]),
                            set([tuple(v) for v in df_node_orient.values]))
        self.assertSetEqual(set([tuple(v) for v in df_edge_pandas.values]),
                            set([tuple(v) for v in df_edge_orient.values]))
        self.assertSetEqual(set(df_node_pandas.index),
                            set(df_node_orient.index))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Convert a graph between Pandas and OrientDB.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import copy
import json

import pandas as pd
import pyorient.otypes
from pyorient.utils import get_hash

from .utils import _find_field_types
from ..utils import byteify, chunks

def as_pandas(nodes=[], edges=[], force_rid=False, deepcopy = True):
    """
    Converts OrientDB Gremlin query results into Pandas DataFrame.

    Parameters
    ----------
    nodes : list of pyorient.otypes.OrientRecord
        OrientDB node query results.
    edges : list of pyorient.otypes.OrientRecord
        OrientDB edge query results.
    force_rid : bool
        If True, always use the OrientDB RID as the index value in the
        returned DataFrame of node data. Otherwise, use the 'id' property as the
        index value if it is present.

    Returns
    -------
    df_node, df_edge : pandas.DataFrame
        DataFrame instances containing query results. The OrientDB class of each
        node and edge is stored in the 'class' column of the corresponding
        DataFrame instance.
    """

    props_list = []
    index = []
    rid_to_id = {}
    for node in nodes:
        # Don't let function alter the original records:
        if deepcopy:
            tmp = copy.deepcopy(node.oRecordData)
        else:
            tmp = node.oRecordData
        props = {}
        for k, v in tmp.items():
            if isinstance(v, pyorient.otypes.OrientBinaryObject):
                continue
            if isinstance(k, str) and k.startswith('_'):
                continue
            if isinstance(v, pyorient.otypes.OrientRecordLink):
                props[k] = get_hash(v)
            elif (isinstance(v,list) and v and
                       isinstance(v[0], pyorient.otypes.OrientRecordLink)):
                props[k] = [get_hash(n) for n in v]
            else:
                props[k] = v
        # props_keys = list(props.keys())
        # for k in props_keys:
        #
        #     # Discard binary objects:
        #     if isinstance(props[k], pyorient.otypes.OrientBinaryObject):
        #         del props[k]
        #
        #
        #     # Replace record links with their corresponding RIDs:
        #     #elif isinstance(props[k], pyorient.otypes.OrientRecordLink):
        #     #    props[k] = props[k].get_hash()
        #     # Remove record links
        #     elif isinstance(props[k], pyorient.otypes.OrientRecordLink):
        #         del props[k]
        #
        #     # Remove list of links
        #     elif (isinstance(props[k],list) and props[k] and
        #           isinstance(props[k][0], pyorient.otypes.OrientRecordLink)):
        #         del props[k]
        #
        #     # Remove properties whose name is a string that starts with '_'; they
        #     # are for special OrientDB purposes:
        #     elif isinstance(k, str) and k.startswith('_'):
        #         del props[k]

        # Save the OrientDB class:
        props['class'] = node._class

        # If the node has an 'id' attribute, use that as the index value if
        # force_rid isn't set:
        if 'id' in props and not force_rid:
            id = props['id']
            del props['id']
        else:
            id = props.get('rid', node._rid)
        index.append(id)
        props_list.append(props)

        rid_to_id[props.get('rid', node._rid)] = id
        # rid_to_id[node._rid] = id
    df_node = pd.DataFrame.from_records(props_list)
    df_node.index = pd.Index(data=index, name='id')

    prop_list = []
    for edge in edges:
        # Don't let function alter the original records:
        if deepcopy:
            tmp = copy.deepcopy(edge.oRecordData)
        else:
            tmp = edge.oRecordData

        props = {k:v for k,v in tmp.items()}
        # Convert record IDs to the IDs assigned to the nodes:
        props['in'] = rid_to_id[props['in'].get_hash()]
        props['out'] = rid_to_id[props['out'].get_hash()]

        # Save the OrientDB class:
        props['class'] = edge._class

        prop_list.append(props)

    # Don't preserve the OrientDB edge IDs:
    df_edge = pd.DataFrame.from_records(prop_list)
    df_edge.index.name = 'id'

    return df_node, df_edge

def orient_to_pandas(client, node_query='', edge_query='', force_rid=False):
    """
    Query OrientDB and return results as Pandas DataFrames.

    Parameters
    ----------
    client : pyorient.orient.OrientDB
        OrientDB interface.
    node_query : str
        Gremlin query that returns a collection of nodes.
    edge_query : str
        Gremlin query that returns a collection of edges.
    force_rid : bool
        If True, always use the OrientDB RID as the index value in the
        returned DataFrame of node data. Otherwise, use the 'id' property as the
        index value if it is present.

    Returns
    -------
    df_node, df_edge : pandas.DataFrame
        DataFrame instances containing query results. The OrientDB class of each
        node and edge is stored in the 'class' column of the corresponding
        DataFrame instance.
    """

    if node_query:
        nodes = client.gremlin(node_query)
    else:
        nodes = []
    if edge_query:
        edges = client.gremlin(edge_query)
    else:
        edges = []
    return as_pandas(nodes, edges, force_rid)

def pandas_to_orient(client, df_node, df_edge):
    """
    Loads Pandas DataFrames into OrientDB database.

    Parameters
    ----------
    client : pyorient.orient.OrientDB
        OrientDB interface.
    df_node, df_edge : pandas.DataFrame
        Tables containing the properties of each node and edge to convert.

    Notes
    -----
    Node IDs are discarded upon creation of the new graph
    """

    assert isinstance(df_node, pd.DataFrame)
    assert isinstance(df_edge, pd.DataFrame)

    N = 10
    id_to_rid = {}
    for chunk in chunks(zip(df_node.index, df_node.to_dict('record')), N):
        cmd_list = []
        id_list = []
        for i, (id, props) in enumerate(chunk):
            if 'class' in props:
                cls = props['class']
                del props['class']
            else:
                cls = 'V'

            # Save original ID:
            assert 'id' not in props
            props['id'] = id

            # Add @fieldTypes field to force proper storage of types:
            ft = _find_field_types(props)
            if ft:
                props['@fieldTypes'] = ft

            id_list.append(id)
            cmd_list.append('let $a%s = create vertex %s content %s;' % \
                            (i, cls, json.dumps(byteify(props))))
        cmd = 'begin;'+''.join(cmd_list)+'commit;'+\
              ('return [%s];' % ','.join(['$a%s' % i for i in range(len(chunk))]))
        rec_list = client.batch(cmd)
        for id, r in zip(id_list, rec_list):
            id_to_rid[id] = r._rid

    for chunk in chunks(zip(df_edge.index, df_edge.to_dict('record')), N):
        for id, props in chunk:
            from_id = props['out']
            to_id = props['in']
            if 'class' in props:
                cls = props['class']
                del props['class']
            else:
                cls = 'E'
            del props['out']
            del props['in']

            # Add @fieldTypes field to force proper storage of types:
            ft = _find_field_types(props)
            if ft:
                props['@fieldTypes'] = ft

            cmd_list = ['create edge %s from %s to %s content %s;' % \
                        (cls, id_to_rid[from_id], id_to_rid[to_id], json.dumps(props))]
            cmd = 'begin;'+''.join(cmd_list)+'commit;'
            client.batch(cmd)



#!/usr/bin/env python

"""
Visualization tools.
"""

import os.path
import tempfile

import matplotlib.pyplot as plt
import networkx as nx

def imdisp(f):
    """
    Display the specified image file using matplotlib.

    Parameters
    ----------
    f : str
        Image file name.
    """
    
    im = plt.imread(f)
    plt.imshow(im)
    plt.axis('off')
    plt.draw()
    return im

def show_pydot(g):
    """
    Display a networkx graph using pydot.

    Parameters
    ----------
    g : networkx.Graph
        Graph to visualize.
    """
    
    fd = tempfile.NamedTemporaryFile()
    fd.close()
    p = nx.to_pydot(g)
    p.write_jpg(fd.name)
    imdisp(fd.name)
    os.remove(fd.name)

def show_pygraphviz(g, prog='dot', graph_attr={}, node_attr={}, edge_attr={}):
    """
    Display a networkx graph using pygraphviz.

    Parameters
    ----------
    g : networkx.Graph
        Graph to visualize.
    prog : str
        Executable for generating the image.
    graph_attr : dict
        Global graph display attributes.
    node_attr : dict
        Global node display attributes.
    edge_attr : dict
        Global edge display attributes.

    See Also
    --------
    pygraphviz
    """
    
    fd = tempfile.NamedTemporaryFile(suffix='.jpg')
    fd.close()
    p = nx.to_agraph(g)
    p.graph_attr.update(graph_attr)
    p.node_attr.update(node_attr)
    p.edge_attr.update(edge_attr)
    p.draw(fd.name, prog=prog)
    imdisp(fd.name)
    os.remove(fd.name)
